import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np

from smoothmax import Smoothmax

from complex import *

def debug_exp():
	import matplotlib.pyplot as plt
	plt.ion()
	plt.clf()
	# torch.abs doesn't work for huge vectors, here is a trick to compute them anyway
	def abs(x):
		d = torch.max(torch.abs(x.real),torch.abs(x.imag))
		return torch.abs(x/d)*d
	x = torch.linspace(0,100,2**12)
	# for s in torch.linspace(0,5,10): plt.plot(x,Exp(2**6,s)(x))
	# for s in torch.linspace(0,5,10): y=Exp(2**6,s)(x*torch.pi*1j) ; plt.plot(x+y.real,y.imag)
	# for s in torch.linspace(0,5,10): y=Exp(2**6,s)(x*torch.pi*1j) ; plt.plot(x,y.angle())
	# for s in torch.linspace(0,5,10): y=Exp(2**6,s)(x*(1.+torch.pi*1j)) ; plt.plot(x,y.abs())
	# for s in torch.linspace(0,5,10): y=Exp(2**6,s)(x*(1.+torch.pi*1j)) ; plt.plot(x,abs(y))
	from debug_tools import show_c
	# show_c(Exp(2**0),[0,5])
	show_c(Exp(2**2,0.1),[-1,7],[-np.pi,np.pi],1)

debug_exp()

def debug_log():
	import matplotlib.pyplot as plt
	plt.ion()
	plt.clf()
	# x = torch.linspace(-1,1,2**12)+0j
	x = torch.linspace(-.01,.01,2**12)+0j
	# for s in torch.linspace(0,5,10): plt.plot(x,C_Log(2**6,s)(x))
	smoothness= 0.5
	# plt.plot(x,C_Log(2**-2,smoothness)(x).real);plt.plot(x,C_Log(2**-2,smoothness)(x).imag)
	plt.plot(x,torch.log(x).real)
	# plt.plot(x,torch.log(x).imag)
	plt.plot(x,C_Log()(x).real)
	# plt.plot(x,C_Log()(x).imag)
	# plt.plot(x,C_Log(2**-10,smoothness)(x).real)
	# plt.plot(x,C_Log(2**-10,smoothness)(x).imag)
	# for s in torch.linspace(0,5,10): y=C_Log(2**-8,s)(x*torch.pi*1j) ; plt.plot(x+y.real,y.imag)
	# for s in torch.linspace(0,5,10): y=C_Log(2**-8,s)(x*torch.pi*1j) ; plt.plot(x,y.angle())
	# for s in torch.linspace(0,5,10): y=C_Log(2**-8,s)(x*(1.+torch.pi*1j)) ; plt.plot(x,y.abs())
	# for s in torch.linspace(0,5,10): y=C_Log(2**-8,s)(x*(1.+torch.pi*1j)) ; plt.plot(x,abs(y))
	from debug_tools import show_c
	# show_c(C_Log(2**-2),[-1,2],[-1,2])
	# show_c(C_Log(),[-1,2],[-1,2])

debug_log()

###############################################################################
### RELU ######################################################################
###############################################################################

def debug_relu():
	import matplotlib.pyplot as plt
	plt.ion()
	plt.clf()
	from debug_tools import show_c
	show_c(C_ReLU())

debug_relu()

###############################################################################
### LINEAR ####################################################################
###############################################################################

def debug_linear():
	import matplotlib.pyplot as plt
	plt.ion()
	plt.clf()
	N=200
	dim_i=700
	dim_o=500
	x = torch.randn(N,dim_i)
	y = torch.randn(N,dim_o)
	def try_step(model,x,y):
		optim = torch.optim.Adam(model.parameters(),lr=1e-2)
		optim.zero_grad()
		torch.mean((model(x)-y).abs()).backward()
		optim.step()
		print(model(x).shape)
	# model=C_Linear_Real(dim_i,dim_o)   ; try_step(model,x,y)   ; plt.hist(torch.flatten(model.A.weight.detach()),128)
	# model=C_Linear_Real(dim_i,dim_o)   ; try_step(model,x+1j,y); plt.hist(torch.flatten(model.A.weight.detach()),32)
	# model=C_Linear_Sigmoid(dim_i,dim_o); try_step(model,x,y)   ; plt.hist(torch.flatten(model.A.detach()),32)
	# model=C_Linear_Sigmoid(dim_i,dim_o); try_step(model,x+1j,y); plt.hist(torch.flatten(model.A.detach()),32)
	# model=C_Linear_Sigmoid(dim_i,dim_o); try_step(model,x,y)   ; plt.hist(torch.flatten(model(x)).detach(),32,density=True)
	# model=C_Linear_Sigmoid(dim_i,dim_o); try_step(model,x,y)   ; plt.hist(torch.flatten(torch.sigmoid(model.A)).detach(),32,density=True)
	model=C_Linear_2_Softmax(dim_i,dim_o); try_step(model,x,y)   ; plt.hist(torch.flatten(model(x)).detach(),32,density=True)
	# plt.hist(torch.randn(7000),32)

debug_linear()





########## C_Linear_Sigmoid
# x = torch.randn(2**16)*4
# plt.clf();plt.hist(F.sigmoid(x),64,density=True)

# plt.clf();plt.hist(torch.flatten(F.sigmoid(torch.randn(dim_i, dim_o)-2.)),32,density=True)
# plt.clf();plt.hist(torch.sum(F.sigmoid(torch.randn(dim_i, dim_o)-6.),axis=0),32,density=True)

# 𝔼(∑ sigmoid(torch.randn(dim_i)-μ) )

# distribution de sigmoid(N(μ,σ))
# distribution N(μ,σ)(x) = exp(-[(x-μ)/σ]²/2) / (σ sqrt(2π))
# definition   sigmoid(x)=

# def normal(x,μ=0,σ=1): return torch.exp(-(((x-μ)/σ)**2) /2 ) / (σ*np.sqrt(2*np.pi))

# def inv_sigmoid(y): return torch.log(y/(1-y))

# plt.clf()
# torch.sum(normal(x))*(x[1]-x[0])

# x = torch.linspace(0,1,2**12)
# torch.sum(normal(inv_sigmoid(x)))*(x[1]-x[0])

# x = torch.linspace(-5,5,2**12)
# plt.clf()
# μ=0
# plt.plot(x,normal(inv_sigmoid(x),μ))
# plt.hist(F.sigmoid(torch.randn(2**14)+μ),128,density=True)


# plt.hist(torch.sum(F.sigmoid(torch.randn(dim_i, dim_o)-6.),axis=0),32,density=True)
# plt.plot(x,normal(inv_sigmoid(x)))



########## C_Linear_2_Softmax

# torch.sigmoid(torch.randn(10,5).fill_diagonal_(1,wrap=True))

# dim_i=2048
# dim_o=15
# A0 = torch.randn(dim_i, dim_o)
# A1 = torch.randn(dim_i, dim_o)
# a1 = torch.zeros(dim_o) # switch on or off the second parameter
# sm = nn.Softmax(0)
# A = sm(A0)
# plt.clf()
# plt.hist(torch.flatten(A),128,density=True)
# for line in A:
# 	print("hi")
# 	line = [x.item() for x in sorted(line)]
# 	plt.hist(line,128,density=True)
# 	None

# plt.yscale("log")
# A = A + torch.sigmoid(a1)*sm(A1) 

########## C_Linear_2_Softmax

